# ----------------------------------------------------------------------------
# Python Wires
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Python Wiring Shell.
"""

from __future__ import absolute_import

from . import _instance



class WiringShell(object):

    """
    Python Wiring Shell
    """

    # Holds the default, per-callable, `min_wirings` and `max_wirings` as well
    # as the default caller/callee call `coupling` mode.
    #
    # Wraps a Wiring Instance, cooperating with it by setting its `coupling`
    # attribute to support call-time caller/callee coupling overriding, while
    # delegating attribute access to expose the Wiring Instance behaviour.

    def __init__(self, min_wirings=None, max_wirings=None, coupling=False):

        if min_wirings is not None and min_wirings <= 0:
            raise ValueError('min_wirings must be positive or None')
        if max_wirings is not None and max_wirings <= 0:
            raise ValueError('max_wirings must be positive or None')
        if min_wirings and max_wirings and min_wirings > max_wirings:
            raise ValueError('max_wirings must be >= min_wirings')

        self._min_wirings = min_wirings
        self._max_wirings = max_wirings
        self._coupling = coupling
        self._wiring_instance = _instance.WiringInstance(self)


    @property
    def min_wirings(self):
        """
        Read-only default minimum wired callables.
        """
        return self._min_wirings


    @property
    def max_wirings(self):
        """
        Read-only default maximum wired callables.
        """
        return self._max_wirings


    @property
    def coupling(self):
        """
        Read-only default call coupling mode.
        """
        return self._coupling


    @property
    def couple(self):
        """
        Caller/callee coupled call context attribute (overrides default).
        """
        self._wiring_instance.coupling = True
        return self._wiring_instance


    @property
    def decouple(self):
        """
        Caller/callee decoupled call context attribute (overrides default).
        """
        self._wiring_instance.coupling = False
        return self._wiring_instance


    def __getattr__(self, name):

        return getattr(self._wiring_instance, name)


    def __getitem__(self, name):

        return self.__getattr__(name)


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Test callables mixin.
"""


from __future__ import absolute_import

from . import helpers



class TestCallablesMixin(object):

    """
    Holds test callables.
    """

    @staticmethod
    def returns_42_callable():

        return 42


    @staticmethod
    def returns_None_callable():

        return None


    THE_EXCEPTION = ValueError('bad value detail')

    def raises_exception_callable(self):

        raise self.THE_EXCEPTION


    def unwire_call(self, callable_):

        self.w.this.unwire(callable_)


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Generic usage test driver mixin.
"""


from __future__ import absolute_import

from . import helpers



class TestWiresUsageMixin(helpers.CallTrackerAssertMixin):

    """
    Drives Wires usage tests requiring mixed class to:
    - Have a Wiring instance at self.w.
    - Allow wiring via self.wire.
    - Allow unwiring via self.unwire.
    """

    def test_wire_call(self):
        """
        Wires the `this` call and calls it.
        Checks the wired callable was called once with the correct arguments.
        """
        tracker = helpers.CallTracker()

        self.w.this.wire(tracker)
        self.w.this()

        self.assert_called(tracker, [
            ((), {},),
        ])


    def test_double_wire_call(self):
        """
        Wires the `this` call twice and calls it.
        Checks the wired callable was called twice with the correct arguments.
        """
        tracker = helpers.CallTracker()

        self.w.this.wire(tracker)
        self.w.this.wire(tracker)
        self.w.this()

        self.assert_called(tracker, [
            ((), {},),
            ((), {},),
        ])


    def test_multi_wired_call(self):
        """
        Wires the `this` call to multiple callables and calls it.
        Checks that each of the wired callables was called once with the correct
        arguments.
        """
        num_wirings = 10

        trackers = [helpers.CallTracker() for _ in range(num_wirings)]

        for tracker in trackers:
            self.w.this.wire(tracker)

        self.w.this()

        for tracker in trackers:
            self.assert_called(tracker, [
                ((), {},),
            ])


    def test_wire_call_unwire_call(self):
        """
        Wires the `this` call to a callable and calls it.
        Checks the wired callable was called once with the correct arguments.
        Unwires the `this` callable and calls it again.
        Checks the wired callable wasn't called again.
        """
        tracker = helpers.CallTracker()

        self.w.this.wire(tracker)
        self.w.this()

        self.assert_single_call_no_args(tracker)

        self.w.this.unwire(tracker)
        self.w.this()

        self.assert_single_call_no_args(tracker)


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Argument passing test driver mixin.
"""


from __future__ import absolute_import

from . import helpers, mixin_test_callables



class _TestWiresArgPassingMixin(mixin_test_callables.TestCallablesMixin,
                                helpers.CallTrackerAssertMixin):

    """
    Argument passing utilization tests for the Wires singleton.
    """

    wire1_args = ()
    wire1_kwargs = {}

    wire2_args = ()
    wire2_kwargs = {}

    call_args = ()
    call_kwargs = {}


    def test_wire_call_args(self):
        """
        Wires the `this` call and calls it.
        Checks the wired callable was called once with the correct arguments.
        """
        tracker = helpers.CallTracker()

        self.w.this.wire(tracker, *self.wire1_args, **self.wire1_kwargs)
        self.w.this(*self.call_args, **self.call_kwargs)
        self.addCleanup(self.unwire_call, tracker)

        self.assertEqual(tracker.call_count, 1, 'call count mismatch')
        self.assertEqual(tracker.call_args, [
            (self.expected_call_args, self.expected_call_kwargs),
        ], 'call argument mismatch')


    def test_double_wire_call_args(self):
        """
        Wires the `this` call and calls it.
        Checks the wired callable was called once with the correct arguments.
        """
        tracker = helpers.CallTracker()

        self.w.this.wire(tracker, *self.wire1_args, **self.wire1_kwargs)
        self.w.this.wire(tracker, *self.wire2_args, **self.wire2_kwargs)
        self.w.this(*self.call_args, **self.call_kwargs)
        self.addCleanup(self.unwire_call, tracker)
        self.addCleanup(self.unwire_call, tracker)

        self.assertEqual(tracker.call_count, 2, 'call count mismatch')
        self.assertEqual(tracker.call_args, [
            (self.expected_call_args, self.expected_call_kwargs),
            (self.expected_2nd_call_args, self.expected_2nd_call_kwargs),
        ], 'call argument mismatch')



class TestWiresNoArgPassingMixin(_TestWiresArgPassingMixin):

    """
    No wire-time args/kwargs, no call-time args/kwargs.
    """

    expected_call_args = ()
    expected_call_kwargs = {}

    expected_2nd_call_args = ()
    expected_2nd_call_kwargs = {}



class TestWiresWireArgPassingMixin(_TestWiresArgPassingMixin):

    """
    Wire-time args, no call-time args/kwargs.
    """

    wire1_args = (1, 2, 3)

    expected_call_args = (1, 2, 3)
    expected_call_kwargs = {}

    expected_2nd_call_args = ()
    expected_2nd_call_kwargs = {}



class TestWiresWireKwargPassingMixin(_TestWiresArgPassingMixin):

    """
    Wire-time kwargs, no call-time args/kwargs.
    """

    wire1_kwargs = dict(a='a', b='b')

    expected_call_args = ()
    expected_call_kwargs = dict(a='a', b='b')

    expected_2nd_call_args = ()
    expected_2nd_call_kwargs = {}



class TestWiresWireArgKwargPassingMixin(_TestWiresArgPassingMixin):

    """
    Wire-time args and kwargs, no call-time args/kwargs.
    """

    wire1_args = (1, 2, 3)
    wire1_kwargs = dict(a='a', b='b')

    expected_call_args = (1, 2, 3)
    expected_call_kwargs = dict(a='a', b='b')

    expected_2nd_call_args = ()
    expected_2nd_call_kwargs = {}



class TestWiresCallArgPassingMixin(_TestWiresArgPassingMixin):

    """
    No wire-time args/kwargs, call-time args.
    """

    call_args = (1, 2, 3)

    expected_call_args = (1, 2, 3)
    expected_call_kwargs = {}

    expected_2nd_call_args = (1, 2, 3)
    expected_2nd_call_kwargs = {}



class TestWiresCallKwargPassingMixin(_TestWiresArgPassingMixin):

    """
    No wire-time args/kwargs, call-time kwargs.
    """

    call_kwargs = dict(a='a', b='b')

    expected_call_args = ()
    expected_call_kwargs = dict(a='a', b='b')

    expected_2nd_call_args = ()
    expected_2nd_call_kwargs = dict(a='a', b='b')



class TestWiresCallArgKwargPassingMixin(_TestWiresArgPassingMixin):

    """
    No wire-time args/kwargs, call-time wire args and kwargs.
    """

    call_args = (1, 2, 3)
    call_kwargs = dict(a='a', b='b')

    expected_call_args = (1, 2, 3)
    expected_call_kwargs = dict(a='a', b='b')

    expected_2nd_call_args = (1, 2, 3)
    expected_2nd_call_kwargs = dict(a='a', b='b')



class TestWiresFullPassingMixin(_TestWiresArgPassingMixin):

    """
    Wire-time args and kwargs, call-time args and kwargs.
    """

    wire1_args = (1, 2, 3)
    wire1_kwargs = dict(a='a', b='b')

    call_args = (4, 5, 6)
    call_kwargs = dict(c='c', d='d')

    expected_call_args = (1, 2, 3, 4, 5, 6)
    expected_call_kwargs = dict(a='a', b='b', c='c', d='d')

    expected_2nd_call_args = (4, 5, 6)
    expected_2nd_call_kwargs = dict(c='c', d='d')



class TestWiresDoubleFullPassingMixin(_TestWiresArgPassingMixin):

    """
    Two wirings with separate args and kwargs each, call time arg and kwargs.
    """

    wire1_args = (1, 2, 3)
    wire1_kwargs = dict(a='a', b='b')

    wire2_args = (4, 5, 6)
    wire2_kwargs = dict(c='c', d='d')

    call_args = (7, 8, 9)
    call_kwargs = dict(e='e', f='f')

    expected_call_args = (1, 2, 3, 7, 8, 9)
    expected_call_kwargs = dict(a='a', b='b', e='e', f='f')

    expected_2nd_call_args = (4, 5, 6, 7, 8, 9)
    expected_2nd_call_kwargs = dict(c='c', d='d', e='e', f='f')


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
API test driver mixin.
"""


from __future__ import absolute_import

from . import mixin_test_callables



class TestWiresAPIMixin(mixin_test_callables.TestCallablesMixin):

    """
    Drives Wires API tests requiring mixed class to:
    - Have a Wiring instance at self.w.
    - Allow wiring via self.wire.
    - Allow unwiring via self.unwire.
    """

    def test_unwired_call_does_not_fail(self):
        """
        Calling an unwired call works. Does nothing, but works.
        """
        self.w.unwired_call()


    def test_wiring_non_callable_raises_value_error(self):
        """
        Wiring a call to a non-callable raises ValueError.
        The exception argument (message):
        - Starts with "argument not callable: ".
        - Contains repr(argument).
        """
        for non_callable in (None, True, 42, 2.3, (), [], {}, set()):

            with self.assertRaises(ValueError) as cm:
                self.w.this.wire(non_callable)

            exception_args = cm.exception.args
            self.assertEqual(len(exception_args), 1)

            msg = exception_args[0]
            self.assertTrue(
                msg.startswith('argument not callable: '),
                'wrong exception message: %r' % (msg,),
            )
            self.assertIn(
                repr(non_callable),
                msg,
                'missing argument repr in message: %r' % (msg,),
            )


    def test_unwiring_non_callable_raises_value_error(self):
        """
        Unwiring a call from a non-callable raises ValueError.
        The exception argument (message):
        - Starts with "argument not callable: ".
        - Contains repr(argument).
        """
        for non_callable in (None, True, 42, 2.3, (), [], {}, set()):

            with self.assertRaises(ValueError) as cm:
                self.w.this.unwire(non_callable)

            exception_args = cm.exception.args
            self.assertEqual(len(exception_args), 1)

            msg = exception_args[0]
            self.assertTrue(
                msg.startswith('argument not callable: '),
                'wrong exception message: %r' % (msg,),
            )
            self.assertIn(
                repr(non_callable),
                msg,
                'missing argument repr in message: %r' % (msg,),
            )


    def test_wiring_callable_works(self):
        """
        Wiring a callable works.
        """
        self.w.this.wire(self.returns_42_callable)
        self.addCleanup(self.unwire_call, self.returns_42_callable)


    def test_wiring_unwiring_works(self):
        """
        Wiring and then unwiring same callable works.
        """
        self.w.this.wire(self.returns_42_callable)
        self.w.this.unwire(self.returns_42_callable)


    def test_unwiring_unknown_callable_raises_value_error(self):
        """
        Unwiring an unknown callable raises a ValueError.
        The exception argument (message):
        - Starts with "unknown function ".
        """
        with self.assertRaises(ValueError) as cm:
            self.w.this.unwire(self.returns_42_callable)

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)

        msg = exception_args[0]
        self.assertTrue(
            msg.startswith('unknown function '),
            'wrong exception message: %r' % (msg,),
        )


    def test_dynamic_name_wire_unwire(self):
        """
        Wiring/unwiring via indexing works.
        """
        name = 'name'
        self.w[name].wire(self.returns_42_callable)
        self.w[name].unwire(self.returns_42_callable)


    def _assert_exception_arg(self, cm, expected):

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], expected)


    def test_wiring_from_instance_fails(self):
        """
        Wiring at the instance level raises RuntimeError.
        """
        with self.assertRaises(RuntimeError) as cm:
            self.w.some_callable.calls_to(self.returns_42_callable)

        self._assert_exception_arg(cm, 'undefined wiring context')


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Global Wires singleton argument passing tests.
"""


from __future__ import absolute_import

import unittest

from . import mixin_test_singleton, mixin_test_args



class TestWiresNoArgPassing(mixin_test_singleton.TestWiresSingletonMixin,
                            mixin_test_args.TestWiresNoArgPassingMixin,
                            unittest.TestCase):

    pass



class TestWiresWireArgPassing(mixin_test_singleton.TestWiresSingletonMixin,
                              mixin_test_args.TestWiresWireArgPassingMixin,
                              unittest.TestCase):

    pass



class TestWiresWireKwargPassing(mixin_test_singleton.TestWiresSingletonMixin,
                                mixin_test_args.TestWiresWireKwargPassingMixin,
                                unittest.TestCase):

    pass



class TestWiresWireArgKwargPassing(mixin_test_singleton.TestWiresSingletonMixin,
                                   mixin_test_args.TestWiresWireArgKwargPassingMixin,
                                   unittest.TestCase):

    pass



class TestWiresCallArgPassing(mixin_test_singleton.TestWiresSingletonMixin,
                              mixin_test_args.TestWiresCallArgPassingMixin,
                              unittest.TestCase):

    pass



class TestWiresCallKwargPassing(mixin_test_singleton.TestWiresSingletonMixin,
                                mixin_test_args.TestWiresCallKwargPassingMixin,
                                unittest.TestCase):

    pass



class TestWiresCallArgKwargPassing(mixin_test_singleton.TestWiresSingletonMixin,
                                   mixin_test_args.TestWiresCallArgKwargPassingMixin,
                                   unittest.TestCase):

    pass



class TestWiresFullPassing(mixin_test_singleton.TestWiresSingletonMixin,
                           mixin_test_args.TestWiresFullPassingMixin,
                           unittest.TestCase):

    pass



class TestWiresDoubleFullPassing(mixin_test_singleton.TestWiresSingletonMixin,
                                 mixin_test_args.TestWiresDoubleFullPassingMixin,
                                 unittest.TestCase):

    pass


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Python Wiring Callable.
"""

from __future__ import absolute_import



class WiringCallable(object):

    """
    Callable with a minimal API.
    Summary:
    - Has zero or more wirings: functions/callables wired to it.
    - Each wiring has optional wire-time arguments.
    - Calling it calls all wired functions/callables, passing them the
      combination of call-time and wire-time arguments.
    """

    def __init__(self, wiring_instance):

        self._wiring_instance = wiring_instance

        # Default min/max_wirings to our instance's.
        self._min_wirings = wiring_instance.min_wirings
        self._max_wirings = wiring_instance.max_wirings

        # Wired (<callable>, <wire-time-args>, <wire-time-kwargs>) tuples.
        self._wirings = []


    @property
    def min_wirings(self):
        """
        Minimum number of allowed wirings. None means no limit.
        """
        return self._min_wirings


    @min_wirings.setter
    def min_wirings(self, value):
        """
        Minimum number of allowed wirings. None means no limit. Must be:
        - A positive integer.
        - Less than or equal to max_wirings
        - Less than or equal to the current number of wirings, if any.
        """
        if value is not None:
            wiring_count = len(self._wirings)
            if value <= 0:
                raise ValueError('min_wirings must be positive or None')
            elif self._max_wirings is not None and value > self._max_wirings:
                raise ValueError('min_wirings must be <= max_wirings')
            elif wiring_count and value > wiring_count:
                raise ValueError('too few wirings')

        self._min_wirings = value


    @property
    def max_wirings(self):
        """
        Maximum number of allowed wirings. None means no limit.
        """
        return self._max_wirings


    @max_wirings.setter
    def max_wirings(self, value):
        """
        Maximum number of allowed wirings. None means no limit. Must be:
        - A positive integer.
        - Greater than or equal to min_wirings
        - Greater than or equal to the current number of wirings, if any.
        """
        if value is not None:
            wiring_count = len(self._wirings)
            if value <= 0:
                raise ValueError('max_wirings must be positive or None')
            elif self._min_wirings is not None and value < self._min_wirings:
                raise ValueError('max_wirings must be >= min_wirings')
            elif wiring_count and value < wiring_count:
                raise ValueError('too many wirings')

        self._max_wirings = value


    def wire(self, function, *args, **kwargs):

        """
        Wires `function` with `args` and `kwargs` as wire-time arguments.
        """

        if not callable(function):
            raise ValueError('argument not callable: %r' % (function,))

        # self._max_wirings can be None, meaning "no limit": comparison ok
        if len(self._wirings) == self._max_wirings:
            raise RuntimeError('max_wirings limit reached')

        self._wirings.append((function, args, kwargs))


    def unwire(self, function):

        """
        Unwires `function`.
        If `function` is wired multiple times, just unwires the first wiring.
        """

        if not callable(function):
            raise ValueError('argument not callable: %r' % (function,))

        # self._min_wirings can be None, meaning "no limit": comparison ok
        if len(self._wirings) == self._min_wirings:
            raise RuntimeError('min_wirings limit reached')

        tuples_to_remove = [v for v in self._wirings if v[0] == function]
        if not tuples_to_remove:
            raise ValueError('unknown function %r' % (function,))
        self._wirings.remove(tuples_to_remove[0])


    @staticmethod
    def calls_to(*_args, **_kwargs):

        """
        Called outside of a wiring action context, which makes no sense, as in
        `WiringShell.<callable>.calls_to(...)`.
        """

        raise RuntimeError('undefined wiring context')


    def __call__(self, *args, **kwargs):

        # Calling with wiring count < `min_wirings`, if set, is an error.
        min_wirings = self._min_wirings
        if min_wirings and len(self._wirings) < min_wirings:
            raise RuntimeError('less than min_wirings wired')

        # Get call coupling behaviour for this call from our WiringInstance and
        # then reset it to its default value to account for correct "default"
        # vs "overridden" call coupling behaviour.
        call_coupling = self._wiring_instance.coupling
        self._wiring_instance.coupling_reset()

        # Will contain (<exception>, <result>) per-wiring tuples.
        call_result = []

        for wired_callable, wire_args, wire_kwargs in self._wirings:
            try:
                combined_args = list(wire_args)
                combined_args.extend(args)
                combined_kwargs = dict(wire_kwargs)
                combined_kwargs.update(kwargs)
                result = wired_callable(*combined_args, **combined_kwargs)
                call_result.append((None, result))
            except Exception as e:
                call_result.append((e, None))
                if call_coupling:
                    raise RuntimeError(*call_result)

        return call_result


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Python Wiring Instance.
"""

from __future__ import absolute_import

from . import _callable



class WiringInstance(object):

    """
    Python Wiring Instance
    """

    # Should be used wrapped in a Wires Shell.

    def __init__(self, wiring_shell):

        # Needed to track parameters like `coupling`.
        self._wiring_shell = wiring_shell

        # Tracks known Callable instances:
        # - Keys are callable names (this instance's dynamic attributes).
        # - Values are WiringCallable objects.
        self._wiring_callables = {}

        # Call coupling behavior is set by the shell, which will be dynamically
        # overridden via its `couple` and `decouple` attributes, that set our
        # `coupled` attribute. Callables check this attribute to determine
        # call-time coupling and *must* call our `coupling_reset` after that
        # to ensure correct "default" vs "overridden" coupling.
        self.coupling = wiring_shell.coupling


    @property
    def min_wirings(self):
        """
        Read-only default minimum wired callables.
        """
        return self._wiring_shell.min_wirings


    @property
    def max_wirings(self):
        """
        Read-only default maximum wired callables.
        """
        return self._wiring_shell.max_wirings


    def coupling_reset(self):
        """
        Resets call coupling behaviour to our shell's default.
        """
        self.coupling = self._wiring_shell.coupling


    def __getattr__(self, name):

        # Called on attribute access, returns an callable object.
        # Either uses a tracked one or creates new one, tracking it.

        try:
            return self._wiring_callables[name]
        except KeyError:
            new_callable = _callable.WiringCallable(self)
            self._wiring_callables[name] = new_callable
            return new_callable


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Wires instance caller/callee custom coupling tests.
"""


from __future__ import absolute_import

import unittest

from wires import Wiring

from . import mixin_test_callables



class _TestWiresCouplingMixin(mixin_test_callables.TestCallablesMixin):

    """
    Caller/callee custom coupling tests for Wires instances.
    """

    def wire_returns_42_callable(self):

        self.w.this.wire(self.returns_42_callable)
        self.addCleanup(self.unwire_call, self.returns_42_callable)


    def wire_raises_exeption_callable(self):

        self.w.this.wire(self.raises_exception_callable)
        self.addCleanup(self.unwire_call, self.raises_exception_callable)


    def wire_three_callables_2nd_one_failing(self):

        self.w.this.wire(self.returns_42_callable)
        self.w.this.wire(self.raises_exception_callable)
        self.w.this.wire(self.returns_None_callable)
        self.addCleanup(self.unwire_call, self.returns_None_callable)
        self.addCleanup(self.unwire_call, self.raises_exception_callable)
        self.addCleanup(self.unwire_call, self.returns_42_callable)



class TestWiresCouplingTrue(_TestWiresCouplingMixin, unittest.TestCase):

    """
    Caller/callee explicit default coupled tests for Wires instances.
    """

    def setUp(self):

        self.w = Wiring(coupling=True)


    def test_wire_default_coupled_call(self):

        self.wire_returns_42_callable()

        result = self.w.this()

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], (None, 42))


    def test_wire_default_coupled_fail(self):

        self.wire_raises_exeption_callable()

        with self.assertRaises(RuntimeError) as cm:
            _ = self.w.this()

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], (self.THE_EXCEPTION, None))
    

    def test_wire_wire_wire_default_coupled_fail(self):

        self.wire_three_callables_2nd_one_failing()

        with self.assertRaises(RuntimeError) as cm:
            _ = self.w.this()

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 2)
        self.assertEqual(exception_args[0], (None, 42))
        self.assertEqual(exception_args[1], (self.THE_EXCEPTION, None))


    def test_wire_coupled_call(self):

        self.wire_returns_42_callable()

        result = self.w.couple.this()

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], (None, 42))


    def test_wire_coupled_fail(self):

        self.wire_raises_exeption_callable()

        with self.assertRaises(RuntimeError) as cm:
            _ = self.w.couple.this()

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], (self.THE_EXCEPTION, None))
    

    def test_wire_wire_wire_coupled_fail(self):

        self.wire_three_callables_2nd_one_failing()

        with self.assertRaises(RuntimeError) as cm:
            _ = self.w.couple.this()

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 2)
        self.assertEqual(exception_args[0], (None, 42))
        self.assertEqual(exception_args[1], (self.THE_EXCEPTION, None))


    def test_wire_decoupled_call(self):

        self.wire_returns_42_callable()

        result = self.w.decouple.this()

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], (None, 42))


    def test_wire_decoupled_fail(self):

        self.wire_raises_exeption_callable()

        result = self.w.decouple.this()

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], (self.THE_EXCEPTION, None))
    

    def test_wire_wire_wire_decoupled_fail(self):

        self.wire_three_callables_2nd_one_failing()

        result = self.w.decouple.this()

        self.assertEqual(len(result), 3)
        self.assertEqual(result[0], (None, 42))
        self.assertEqual(result[1], (self.THE_EXCEPTION, None))
        self.assertEqual(result[2], (None, None))



class TestWiresCouplingFalse(_TestWiresCouplingMixin, unittest.TestCase):

    """
    Caller/callee explicit default decoupled tests for Wires instances.
    """

    def setUp(self):

        self.w = Wiring(coupling=False)


    def test_wire_default_decoupled_call(self):

        self.wire_returns_42_callable()

        result = self.w.this()

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], (None, 42))


    def test_wire_default_decoupled_fail(self):

        self.wire_raises_exeption_callable()

        result = self.w.this()

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], (self.THE_EXCEPTION, None))
    

    def test_wire_wire_wire_default_decoupled_fail(self):

        self.wire_three_callables_2nd_one_failing()

        result = self.w.this()

        self.assertEqual(len(result), 3)
        self.assertEqual(result[0], (None, 42))
        self.assertEqual(result[1], (self.THE_EXCEPTION, None))
        self.assertEqual(result[2], (None, None))


    def test_wire_coupled_call(self):

        self.wire_returns_42_callable()

        result = self.w.couple.this()

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], (None, 42))


    def test_wire_coupled_fail(self):

        self.wire_raises_exeption_callable()

        with self.assertRaises(RuntimeError) as cm:
            _ = self.w.couple.this()

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 1)
        self.assertEqual(exception_args[0], (self.THE_EXCEPTION, None))
    

    def test_wire_wire_wire_coupled_fail(self):

        self.wire_three_callables_2nd_one_failing()

        with self.assertRaises(RuntimeError) as cm:
            _ = self.w.couple.this()

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 2)
        self.assertEqual(exception_args[0], (None, 42))
        self.assertEqual(exception_args[1], (self.THE_EXCEPTION, None))


    def test_wire_decoupled_call(self):

        self.wire_returns_42_callable()

        result = self.w.decouple.this()

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], (None, 42))


    def test_wire_decoupled_fail(self):

        self.wire_raises_exeption_callable()

        result = self.w.decouple.this()

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], (self.THE_EXCEPTION, None))
    

    def test_wire_wire_wire_decoupled_fail(self):

        self.wire_three_callables_2nd_one_failing()

        result = self.w.decouple.this()

        self.assertEqual(len(result), 3)
        self.assertEqual(result[0], (None, 42))
        self.assertEqual(result[1], (self.THE_EXCEPTION, None))
        self.assertEqual(result[2], (None, None))


# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
# Python Wires Tests
# ----------------------------------------------------------------------------
# Copyright (c) Tiago Montes.
# See LICENSE for deatils.
# ----------------------------------------------------------------------------

"""
Caller/callee coupling test driver mixin.
"""


from __future__ import absolute_import

from . import mixin_test_callables



class TestCallerCalleeCouplingMixin(mixin_test_callables.TestCallablesMixin):

    """
    Drives Wires caller/callee coupling tests requiring mixed class to:
    - Have a Wiring instance at self.w.
    - Allow wiring via self.wire.
    - Allow unwiring via self.unwire.
    """

    def test_wire_default_decoupled_call(self):
        """
        Default uncoupled test: return a list of (<exception>, <result>), per-
        callee tuple.
        """
        self.w.this.wire(self.returns_42_callable)
        self.addCleanup(self.unwire_call, self.returns_42_callable)
        result = self.w.this()

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], (None, 42))


    def test_wire_wire_wire_default_decoupled_call(self):
        """
        Default uncoupled test: return a list of (<exception>, <result>), per-
        callee tuple.
        """
        self.w.this.wire(self.returns_42_callable)
        self.w.this.wire(self.raises_exception_callable)
        self.w.this.wire(self.returns_None_callable)
        self.addCleanup(self.unwire_call, self.returns_None_callable)
        self.addCleanup(self.unwire_call, self.raises_exception_callable)
        self.addCleanup(self.unwire_call, self.returns_42_callable)
        result = self.w.this()

        self.assertEqual(len(result), 3)
        self.assertEqual(result[0], (None, 42))
        self.assertEqual(result[1], (self.THE_EXCEPTION, None))
        self.assertEqual(result[2], (None, None))


    def test_wire_force_coupled_call(self):
        """
        Wire a callable, call it forcing coupling.
        """
        self.w.this.wire(self.returns_42_callable)
        self.addCleanup(self.unwire_call, self.returns_42_callable)
        result = self.w.couple.this()

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], (None, 42))


    def test_wire_wire_wire_coupled_call(self):
        """
        Wire a callable three times: the second one raises an exception.
        """
        self.w.this.wire(self.returns_42_callable)
        self.w.this.wire(self.raises_exception_callable)
        self.w.this.wire(self.returns_None_callable)
        self.addCleanup(self.unwire_call, self.returns_None_callable)
        self.addCleanup(self.unwire_call, self.raises_exception_callable)
        self.addCleanup(self.unwire_call, self.returns_42_callable)

        with self.assertRaises(RuntimeError) as cm:
            self.w.couple.this()

        exception_args = cm.exception.args
        self.assertEqual(len(exception_args), 2)
        self.assertEqual(exception_args[0], (None, 42))
        self.assertEqual(exception_args[1], (self.THE_EXCEPTION, None))


    def test_wire_force_decoupled_call(self):
        """
        Wire a callable, call it forcing coupling.
        """
        self.w.this.wire(self.returns_42_callable)
        self.addCleanup(self.unwire_call, self.returns_42_callable)
        result = self.w.decouple.this()

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], (None, 42))


    def test_wire_wire_wire_decoupled_call(self):
        """
        Wire a callable three times: the second one raises an exception.
        """
        self.w.this.wire(self.returns_42_callable)
        self.w.this.wire(self.raises_exception_callable)
        self.w.this.wire(self.returns_None_callable)
        self.addCleanup(self.unwire_call, self.returns_None_callable)
        self.addCleanup(self.unwire_call, self.raises_exception_callable)
        self.addCleanup(self.unwire_call, self.returns_42_callable)

        result = self.w.decouple.this()

        self.assertEqual(len(result), 3)
        self.assertEqual(result[0], (None, 42))
        self.assertEqual(result[1], (self.THE_EXCEPTION, None))
        self.assertEqual(result[2], (None, None))


# ----------------------------------------------------------------------------
